@@ -2,7 +2,7 @@ import enum
 import math, re
 
 from time import sleep
-from typing import Callable, Optional
+from typing import Callable, Optional, Tuple
 from collections import OrderedDict
 from copy import deepcopy
 
@@ -13,12 +13,18 @@ from fontTools.pens.transformPen import TransformPen
 from fontTools.pens.reverseContourPen import ReverseContourPen
 
 from fontPens.flattenPen import FlattenPen
-from drafting.geometry import Atom, Point, Line, Edge, Rect, align
+from drafting.geometry import Atom, Point, Line, Edge, Rect, Curve, align
 from drafting.color import normalize_color
 from drafting.sh import SH_UNARY_SUFFIX_PROPS, sh, SHContext
 
 from drafting.pens.misc import BooleanOp, calculate_pathop, ExplodingPen, SmoothPointsPen
 
+from drafting.pens.outlinepen import OutlinePen
+from drafting.pens.translationpen import TranslationPen, polarCoord
+
+from drafting.beziers import CurveCutter, splitCubicAtT
+
+from drafting.interpolation import norm
 
 class DraftingPen(RecordingPen, SHContext):
     """Fluent subclass of RecordingPen"""
@@ -37,6 +43,10 @@ class DraftingPen(RecordingPen, SHContext):
         self._visible = True
         self._parent = None
         self._last = None
+        self._alpha = 1
+
+        self._typographic = False
+        self.glyphName = None
 
         self._current_attr_tag = "default"
         self.clearAttrs()
@@ -60,13 +70,16 @@ class DraftingPen(RecordingPen, SHContext):
         s = f"{type(self).__name__}<"
         if self._tag:
             s += self._tag + ":"
-        s += f"{len(self.value)}mvs:"
-        if self.value[-1][0] == "closePath":
-            s += "closed"
-        elif self.value[-1][0] == "endPath":
-            s += "end"
+        if len(self.value) == 0:
+            s += "(((empty)))"
         else:
-            s += "open"
+            s += f"{len(self.value)}mvs:"
+            if self.value[-1][0] == "closePath":
+                s += "closed"
+            elif self.value[-1][0] == "endPath":
+                s += "end"
+            else:
+                s += "open"
         s += "/>"
         return s
     
@@ -124,6 +137,13 @@ class DraftingPen(RecordingPen, SHContext):
         else:
             return self.bounds()
     
+    def addFrame(self, frame, typographic=False, passthru=False):
+        """Add a new frame to the DATPen, replacing any old frame. Passthru ignored, there for compatibility"""
+        self._frame = frame
+        if typographic:
+            self.typographic = True
+        return self
+    
     def unended(self):
         if len(self.value) == 0:
             return True
@@ -149,6 +169,7 @@ class DraftingPen(RecordingPen, SHContext):
             start = self.value[0][1][-1]
         except:
             start = None
+        #print("SH", s, self.defs)
         res = sh(s, self, subs={"¬":self._last, "§":start, **subs})
         if res[0] == "∫":
             res = [self.single_pen_class().gs(res[1:])]
@@ -191,6 +212,9 @@ class DraftingPen(RecordingPen, SHContext):
                 getattr(self, move)(_e)
             elif isinstance(_e, Rect):
                 self.rect(_e)
+            elif isinstance(_e, Curve):
+                _, b, c, d = _e
+                self.curveTo(b, c, d)
             elif isinstance(_e, str):
                 getattr(self, _e)()
             elif _e[0][0] == "∑":
@@ -201,7 +225,10 @@ class DraftingPen(RecordingPen, SHContext):
                     else:
                         raise Exception("unrecognized macro '" + macro + "'")
             elif len(_e) >= 3:
-                self.boxCurveTo(*_e)
+                if len(_e) >= 5:
+                    self.interpCurveTo(*_e)
+                else:
+                    self.boxCurveTo(*_e)
 
         locals = {}
         mvs = [moves[0]]
@@ -241,7 +268,10 @@ class DraftingPen(RecordingPen, SHContext):
         return self
 
     def lineTo(self, p1):
-        super().lineTo(p1)
+        if len(self.value) == 0:
+            super().moveTo(p1)
+        else:
+            super().lineTo(p1)
         self._last = p1
         return self
 
@@ -276,7 +306,7 @@ class DraftingPen(RecordingPen, SHContext):
         if hasattr(pen, "pens"):
             for p in pen:
                 self.record(p)
-        if pen:
+        elif pen:
             pen.replay(self)
         return self
     
@@ -366,8 +396,20 @@ class DraftingPen(RecordingPen, SHContext):
             self.lineTo(pt)
         self.closePath()
         return self
+
+    def interpCurveTo(self, p1, f1, p2, f2, to, inset=0):
+        a = Point(self.value[-1][-1][-1])
+        d = Point(to)
+        pl = Line(p1, p2).inset(inset)
+        b = Line(a, pl.start).t(f1/100)
+        c = Line(d, pl.end).t(f2/100)
+        return self.curveTo(b, c, d)
     
-    def boxCurveTo(self, point, factor, pt, po=(0, 0), mods={}):
+    def boxCurveTo(self, point, factor, pt, po=(0, 0), mods={}, flatten=False):
+        if flatten:
+            self.lineTo(pt)
+            return self
+        
         a = Point(self.value[-1][-1][-1])
         d = Point(pt)
         box = Rect.FromMnMnMxMx([min(a.x, d.x), min(a.y, d.y), max(a.x, d.x), max(a.y, d.y)])
@@ -420,16 +462,28 @@ class DraftingPen(RecordingPen, SHContext):
         return self
     
     def copy(self, with_data=False):
-        dp = type(self)(self)
+        dp = self.single_pen_class()
+        self.replay(dp)
+        for tag, attrs in self.attrs.items():
+            dp.attr(tag, **attrs)
+        
+        dp.glyphName = self.glyphName
+        dp.defs = self.defs
+
         if with_data:
-            dp._frame = self._frame
-            dp.defs = self.defs # necessary to copy this and not pass by ref?
-        else:
-            dp.defs = self.defs
+            dp.data = self.data
+            if dp._frame:
+                dp._frame = self._frame
+            if hasattr(self, "macros"):
+                dp.macros = self.macros
+            if self.typographic:
+                dp.typographic = True
         return dp
     
     def cast(self, _class, *args):
         """Quickly cast to a (different) subclass."""
+        #if hasattr(self, "pens"):
+        #    return _class(self.pens)
         res = _class(self, *args)
         res.attrs = deepcopy(self.attrs)
         return res
@@ -509,6 +563,14 @@ class DraftingPen(RecordingPen, SHContext):
     def offset_y(self, y):
         return self.translate(0, y)
     
+    def zero_translate(self, th=1, tv=0):
+        x, y, _, _ = self.ambit(th=th, tv=tv)
+        self.translate(-x, -y)
+        return self
+    
+    def center_on_point(self, rect, pt, interp=1):
+        return self.translate(norm(interp, 0, rect.w/2-pt[0]), norm(interp, 0, rect.h/2-pt[1]))
+    
     def skew(self, x=0, y=0, point=None):
         t = Transform()
         if not point:
@@ -614,7 +676,94 @@ class DraftingPen(RecordingPen, SHContext):
             self.repeat(times-1)
         return self
     
+    # Iteration-manipulation
+    
+    def take(self, slice):
+        self.value = self.value[slice]
+        return self
+    
+    def take_curve(self, idx):
+        a = self.value[idx-1][-1][-1]
+        b, c, d = self.value[idx][-1]
+        return Curve(a, b, c, d)
+
+    def ups(self):
+        "Convert this single pen into a collection of pens, with one-pen in the collection (this pen)"
+        dps = self.multi_pen_class()
+        dps.append(self.copy())
+        return dps
+    
+    def collapse(self, levels=100, onself=False):
+        return self.multi_pen_class([self])
+    
+    def mod_pt(self, vidx, pidx, fn):
+        pt = Point(self.value[vidx][-1][pidx])
+        if callable(fn):
+            res = fn(pt)
+        else:
+            res = pt.offset(*fn)
+        try:
+            self.value[vidx][-1][pidx] = res
+        except TypeError:
+            self.pvl()
+            self.value[vidx][-1][pidx] = res
+        return self
+    
+    def mod_pts(self, rect, fn):
+        self.map_points(fn, lambda p: p.inside(rect))
+        return self
+    
+    # Contour manipulation
+
+    def mod_contour(self, contour_index, mod_fn):
+        exploded = self.explode()
+        mod_fn(exploded[contour_index])
+        self.value = exploded.implode().value
+        return self
+    
+    def filter_contours(self, filter_fn):
+        exploded = self.explode()
+        keep = []
+        for idx, c in enumerate(exploded):
+            if filter_fn(idx, c):
+                keep.append(c)
+        self.value = self.multi_pen_class(keep).implode().value
+        return self
+    
+    def slicec(self, contour_slice):
+        self.value = self.multi_pen_class(self.explode()[contour_slice]).implode().value
+        return self
+    
     # Iterating
+
+    def map(self, fn:Callable[[int, str, list], Tuple[str, list]]):
+        for idx, (mv, pts) in enumerate(self.value):
+            self.value[idx] = fn(idx, mv, pts)
+        return self
+    
+    def filter(self, fn:Callable[[int, str, list], bool]):
+        vs = []
+        for idx, (mv, pts) in enumerate(self.value):
+            if fn(idx, mv, pts):
+                vs.append((mv, pts))
+        self.value = vs
+        return self
+    
+    def map_points(self, fn, filter_fn=None):
+        idx = 0
+        for cidx, c in enumerate(self.value):
+            move, pts = c
+            pts = list(pts)
+            for pidx, p in enumerate(pts):
+                x, y = p
+                if filter_fn and not filter_fn(p):
+                    continue
+                result = fn(idx, x, y)
+                if result:
+                    pts[pidx] = result
+                idx += 1
+            self.value[cidx] = (move, pts)
+        return self
     
     def walk(self, callback:Callable[["DraftingPen", int, dict], None], depth=0, visible_only=False, parent=None):
         if visible_only and not self._visible:
@@ -672,13 +821,16 @@ class DraftingPen(RecordingPen, SHContext):
         max_ang = max([l.ang for l in lines])
         #for idx, l in enumerate(lines):
         #    print(idx, ">", l.ang, min_ang, max_ang)
-        xs = [l for l in lines if math.isclose(l.ang, min_ang)]
-        ys = [l for l in lines if math.isclose(l.ang, max_ang)]
+        xs = [l for l in lines if l.ang < 0.25 or l.ang > 2.5]
+        ys = [l for l in lines if 1 < l.ang < 2]
 
         if len(ys) == 2 and len(xs) < 2:
             xs = [l for l in lines if l not in ys]
         elif len(ys) < 2 and len(xs) == 2:
             ys = [l for l in lines if l not in xs]
+        
+        #for l in ys:
+        #    print(l.ang)
 
         #print(len(xs), len(ys))
         #print("--------------------")
@@ -692,10 +844,11 @@ class DraftingPen(RecordingPen, SHContext):
     def avg(self):
         self.pvl()
         pts = []
-        for mv, pts in self.value:
-            if len(pts) > 0:
-                pts.extend(pts)
+        for _, _pts in self.value:
+            if len(_pts) > 0:
+                pts.extend(_pts)
         n = len(pts)
+        #print("AVG", self, self.value)
         return Point(
             sum([p.x for p in pts])/n,
             sum([p.y for p in pts])/n)
@@ -754,6 +907,17 @@ class DraftingPen(RecordingPen, SHContext):
         n, s, e, w = self.nsew()
         return n.interp(0.5, s.reverse())
     
+    def edge(self, e):
+        e = e.lower()
+        if e == "n":
+            return self.en
+        elif e == "s":
+            return self.es
+        elif e == "e":
+            return self.ee
+        elif e == "w":
+            return self.ew
+    
     def shprop(self, s):
         if s in SH_UNARY_SUFFIX_PROPS:
             return SH_UNARY_SUFFIX_PROPS[s]
@@ -786,6 +950,10 @@ class DraftingPen(RecordingPen, SHContext):
     def define(self, *args, **kwargs):
         return self.context_record("$", "defs", None, *args, **kwargs)
     
+    def declare(self, *whatever):
+        # TODO do something with what's declared somehow?
+        return self
+    
     def macro(self, **kwargs):
         for k, v in kwargs.items():
             self.macros[k] = v
@@ -813,7 +981,9 @@ class DraftingPen(RecordingPen, SHContext):
         For simple take-one callback functions in a chain
         """
         if fn:
-            fn(self, *args)
+            res = fn(self, *args)
+            if isinstance(res, DraftingPen):
+                return res
         return self
     
     def replace(self, fn:Callable[["DraftingPen"], None], *args):
@@ -825,9 +995,16 @@ class DraftingPen(RecordingPen, SHContext):
     def cond(self, condition, if_true: Callable[["DraftingPen"], None], if_false=Callable[["DraftingPen"], None]):
         # TODO make if_false optional
         if condition:
-            if_true(self)
+            if callable(if_true):
+                if_true(self)
+            else:
+                self.gs(if_true)
         else:
-            if_false(self)
+            if if_false:
+                if callable(if_false):
+                    if_false(self)
+                else:
+                    self.gs(if_false)
         return self
 
     # BOOLEAN OPERATIONS
@@ -923,6 +1100,14 @@ class DraftingPen(RecordingPen, SHContext):
         self._current_attr_tag = was_tag
         return self
     
+    def calc_alpha(self):
+        a = self._alpha
+        p = self._parent
+        while p:
+            a = a * p._alpha
+            p = p._parent
+        return a
+    
     def v(self, v):
         self.visible(bool(v))
         return self
@@ -991,4 +1176,168 @@ class DraftingPen(RecordingPen, SHContext):
                         if _p:
                             yield _p
                 else:
-                    yield pen
\ No newline at end of file
+                    yield pen
+    
+    # Fun pen manipulations
+
+    def outline(self, offset=1, drawInner=True, drawOuter=True, cap="square"):
+        """AKA expandStroke"""
+        op = OutlinePen(None, offset=offset, optimizeCurve=True, cap=cap)
+        self.replay(op)
+        op.drawSettings(drawInner=drawInner, drawOuter=drawOuter)
+        g = op.getGlyph()
+        p = self.single_pen_class()
+        g.draw(p)
+        self.value = p.value
+        return self
+    
+    ol = outline
+    
+    def project(self, angle, width):
+        offset = polarCoord((0, 0), math.radians(angle), width)
+        self.translate(offset[0], offset[1])
+        return self
+
+    def castshadow(self, angle=-45, width=100, ro=1, fill=1):
+        out = self.single_pen_class()
+        tp = TranslationPen(out, frontAngle=angle, frontWidth=width)
+        self.replay(tp)
+        if fill:
+            out.record(self.copy().project(angle, width))
+        if ro:
+            out.removeOverlap()
+        self.value = out.value
+        return self
+
+    def grow(self, outline=10):
+        out = self.copy().outline(outline)
+        return self.record(out.reverse())
+    
+    def gridlines(self, rect, x=20, y=None, absolute=False):
+        """Construct a grid in the pen using `x` and (optionally) `y` subdivisions"""
+        xarg = x
+        yarg = y or x
+        if absolute:
+            x = int(rect.w / xarg)
+            y = int(rect.h / yarg)
+        else:
+            x = xarg
+            y = yarg
+        
+        for _x in rect.subdivide(x, "minx"):
+            if _x.x > 0 and _x.x > rect.x:
+                self.line([_x.point("NW"), _x.point("SW")])
+        for _y in rect.subdivide(y, "miny"):
+            if _y.y > 0 and _y.y > rect.y:
+                self.line([_y.point("SW"), _y.point("SE")])
+        return self.f(None).s(0, 0.1).sw(3)
+    
+    # Some curvy/bendy things
+
+    def subsegment(self, start=0, end=1):
+        """Return a subsegment of the pen based on `t` values `start` and `end`"""
+        cc = CurveCutter(self)
+        start = 0
+        end = end * cc.calcCurveLength()
+        pv = cc.subsegment(start, end)
+        self.value = pv
+        return self
+    
+    def point_t(self, t=0.5):
+        """Get point value for time `t`"""
+        cc = CurveCutter(self)
+        start = 0
+        tv = t * cc.calcCurveLength()
+        p, tangent = cc.subsegmentPoint(start=0, end=tv)
+        return p, tangent
+    
+    def split_t(self, t=0.5):
+        a = self.value[0][-1][0]
+        b, c, d = self.value[-1][-1]
+        return splitCubicAtT(a, b, c, d, t)
+    
+    def add_pt_t(self, cuidx, t):
+        cidx = 0
+        insert_idx = -1
+        c1, c2 = None, None
+
+        for idx, (mv, pts) in enumerate(self.value):
+            if mv == "curveTo":
+                if cidx == cuidx:
+                    insert_idx = idx
+                    a = self.value[idx-1][-1][-1]
+                    b, c, d = pts
+                    c1, c2 = splitCubicAtT(a, b, c, d, t)
+                cidx += 1
+        
+        if c2:
+            self.value[insert_idx] = ("curveTo", c1[1:])
+            self.value.insert(insert_idx+1, ("curveTo", c2[1:]))
+        return self
+    
+    def length(self, t=1):
+        """Get the length of the curve for time `t`"""
+        cc = CurveCutter(self)
+        start = 0
+        tv = t * cc.calcCurveLength()
+        return tv
+
+    def bend(self, curve, tangent=True):
+        cc = CurveCutter(curve)
+        ccl = cc.length
+        dpl = self.bounds().point("SE").x
+        xf = ccl/dpl
+        def bender(x, y):
+            p, tan = cc.subsegmentPoint(end=x*xf)
+            px, py = p
+            if tangent:
+                a = math.sin(math.radians(180+tan)) * y
+                b = math.cos(math.radians(180+tan)) * y
+                return (px+a, py+b)
+                #return (px, y+py)
+            else:
+                return (px, y+py)
+        return self.nonlinear_transform(bender)
+    
+    def bend2(self, curve, tangent=True, offset=(0, 1)):
+        bw = self.bounds().w
+        a = curve.value[0][-1][0]
+        b, c, d = curve.value[1][-1]
+        def bender(x, y):
+            c1, c2 = splitCubicAtT(a, b, c, d, offset[0] + (x/bw)*offset[1])
+            _, _a, _b, _c = c1
+            if tangent:
+                tan = math.degrees(math.atan2(_c[1] - _b[1], _c[0] - _b[0]) + math.pi*.5)
+                ax = math.sin(math.radians(90-tan)) * y
+                by = math.cos(math.radians(90-tan)) * y
+                return _c[0]+ax, (y+_c[1])+by
+            return _c[0], y+_c[1]
+        return self.nonlinear_transform(bender)
+    
+    def nonlinear_transform(self, fn):
+        for idx, (move, pts) in enumerate(self.value):
+            if len(pts) > 0:
+                _pts = []
+                for _pt in pts:
+                    x, y = _pt
+                    _pts.append(fn(x, y))
+                self.value[idx] = (move, _pts)
+        return self
+    
+    nlt = nonlinear_transform
+    
+    def bend3(self, curve, tangent=False, offset=(0, 1)):
+        a = curve.value[0][-1][0]
+        b, c, d = curve.value[1][-1]
+        bh = self.bounds().h
+        
+        def bender(x, y):
+            c1, c2 = splitCubicAtT(a, b, c, d, offset[0] + (y/bh)*offset[1])
+            _, _a, _b, _c = c1
+            if tangent:
+                tan = math.degrees(math.atan2(_c[1] - _b[1], _c[0] - _b[0]) + math.pi*.5)
+                ax = math.sin(math.radians(90-tan)) * y
+                by = math.cos(math.radians(90-tan)) * y
+                return x+_c[0]+ax, (y+_c[1])+by
+            return x+_c[0], _c[1]
+        return self.nonlinear_transform(bender)
\ No newline at end of file
